import 'package:rapidfire/Dynamic.dart';
import 'package:rapidfire/uiux/Api.dart';
import 'package:rapidfire/uiux/WidgetsRF.dart';
import 'package:sqflite/sqflite.dart';
import 'dart:io' as io;
import 'package:path/path.dart';
import 'package:path_provider/path_provider.dart';

import 'Annotation.dart';
import 'Config.dart';
import 'GlobalRF.dart';
import 'GlobalScopeRF.dart' as g;
import 'GlobalScopeRF.dart';
import 'UtilRF.dart';

class Db {
  //Configuration _configuration = Configuration();
  static Database _database;

  Db(Configuration configuration) {
    //_configuration = configuration;
  }

  Future<Database> get database async {
    if (_database != null) return _database;
    _database = await initDatabase();
    return _database;
  }

  initDatabase() async {
    io.Directory documentDirectory = await getApplicationDocumentsDirectory();
    String path = join(documentDirectory.path, 'student.db');
    var db = await openDatabase(path, version: 1, onCreate: _onCreate);
    return db;
  }

  _onCreate(Database db, int version) async {
    await db.execute(createTable().first);
  }

  get(Object model, String sql, [List<dynamic> values, Function onSuccess]) async {
    List<Object> modelList = List();
    var db = await database;
    List<Map> result = await db.rawQuery(sql, values);
    result.forEach((element) {
      var newModel = Dynamic(model).newClass();
      element.forEach((key, value) {
        Dynamic(newModel).set(key, value);
      });
      modelList.add(newModel);
    });
    if (onSuccess != null) onSuccess(modelList);
    return modelList;
  }

  Future<Object> save(Object model) async {
    var pk = Dynamic(model).getAnnotationData(Key);
    var value = Dynamic(model).get(pk.first['field']);
    if (value == null)
      add(model);
    else
      update(model, '${pk.first['field']} = ?', [value]);
  }

  Future<Object> add(Object model) async {
    var db = await database;
    Dynamic dyn = Dynamic(model);
    var data = dyn.modelToDataMap();
    var id = await db.insert(model.runtimeType.toString(), data);
    return model;
  }

  Future<int> update(Object model, String where, List<dynamic> args) async {
    var data = Dynamic(model).modelToDataMap();
    var db = await database;
    return await db.update(model.runtimeType.toString(), data, where: where, whereArgs: args,);
  }

  Future<int> delete(Object model, int id, [List<Function> validators]) async {
    validators = validators ?? Libs.config.model.validators;
    var isValidValidForDeletion = true;
    var db = await database;
    var pk = Dynamic(model).getAnnotationData(Key);
    validators?.forEach((element) {
      isValidValidForDeletion = element.call();
      if (!isValidValidForDeletion) return;
    });
    if (isValidValidForDeletion)
      return await db.delete('student', where: '${pk.first['field']} = ?', whereArgs: [id],);
    else
      return Future.value(0);
  }

  Future close() async {
    var db = await database;
    db.close();
  }

  List<String> createTable() {
    List<String> createSql = new List();
    var sql = "";
    var pkSql = "";
    var pkFieldName = "";
    var pkFieldType = "";
    var pkFieldAuto = "";
    var models = g.modelList.Get();

    models.forEach((k, v) {
      var annotation = Dynamic(v)
          .getAnnotationData(Key)
          .first;

      if (annotation != null) {
        pkFieldName = annotation["field"];
        pkFieldType = annotation["type"].toString();
        pkFieldAuto = annotation["data"]["auto"] ? "AUTOINCREMENT" : "";
        pkSql = "${pkFieldName} ${getSqlFieldType(pkFieldType)} PRIMARY KEY ${pkFieldAuto}";
      }
      Dynamic(v).modelMembers.forEach((k1, v1) {
        if (pkFieldName != k1) {
          sql = sql + k1 + " " + getSqlFieldType(v1.toString()) + ", ";
        }
      });
      sql = sql.trim().substring(0, sql.length - 2);
      createSql.add("CREATE TABLE ${k}(${pkSql},${sql});");
    });
    return createSql;
  }

  String getSqlFieldType(String s) {
    String type = "";
    if (s.contains("String")) {
      type = " TEXT";
    } else if (s.contains("long")) {
      type = " INTEGER";
    } else if (s.contains("byte[]")) {
      type = " BLOB";
    } else if (s.contains("int")) {
      type = " INTEGER";
    } else {
      type = "";
    }
    return type;
  }

  getPrimaryKeyName(Object model) {
    var pk = Dynamic(model).getAnnotationData(Key);
    return pk.first['field'];
  }

  getPrimaryKeyValue(Object model) {
    var _model = Dynamic(model);
    var pk = _model.getAnnotationData(Key);
    var pkName = pk.first['field'];
    var pkValue = _model.get(pkName.toString());
    return pkValue;
  }
}


//Type typeOfElementsInList<T>(List<T> e) => T;

enum DataSourceType { local, remote, memory }

class DataSourceRF<T> {
  List<T> itemList; // No Use ????
  String sql;
  DataSourceType type;
  Function stateUpdateFn;
  Function readListItemFn;
  Function deleteListItemFn;
  bool isLoading = false;

  DataSourceRF({this.itemList, this.sql, this.type = DataSourceType.local}) {
    var modelType = typeOfElementsInList(itemList);
    this.sql = 'SELECT * FROM ${modelType.toString()} ORDER BY 1 DESC';
  }

  readData(String _sql, int pageNo) async {
    if (!isLoading) {
      isLoading = true;

      var data;
      Api api = Api();

      var model = typeOfElementsInList(itemList);

      if (type == DataSourceType.local) {
        data = await rf.db.get(model, _sql);
      } else {
        data = await api.fetch(model, pageNo: pageNo);
      }

      if (stateUpdateFn != null) stateUpdateFn(data, pageNo);
      isLoading = false;
      PageRF.dataSource = null;
    }
  }

  refresh() {
    readData(sql, 0);
  }

  delete(int index) {
    if (deleteListItemFn != null) deleteListItemFn(index);
  }

  getList() {
    if (readListItemFn != null) return readListItemFn();
  }
}